
#include <stdio.h>
#include <stdlib.h>
#include <string.h>     /* for strlen */

#include <fcntl.h>      /* for open */
#include <unistd.h>     /* for write */

// #include <time.h>       /* asctime, gmtime */

#define TRUE    1
#define FALSE   0
#define BOOLEAN int

#define TEXT = 0
#define NUMBER = 1
#define CHAR = 2

struct t_record {
        int id;
        int option;
        int type;
        void *data;
};

void digit_hexprint(unsigned char character)   {
    printf("%02x ", character);
}

void bitstream_hexprint(unsigned char *buffer, int length)   {
    int i;
    for (i = 0; i < length; i++)  {
        digit_hexprint(buffer[i]);
    }
    printf("\n");
}

unsigned long convert_bitstream_to_number(unsigned char *bitstream) {
    int i;
    unsigned long result = 0;
    unsigned long factor = 1;
    for (i = 0; i < 4; i++) {
        result += factor * bitstream[i];
        factor *= 256;
    }
    return result;
}

BOOLEAN isTextfield(int id, int option) {
        switch (id) {
                case 1 : 
                case 2 : 
                case 3 :
                case 4 :
                case 5 :
                case 8 :
                case 10 :
                case 11 :
                case 15 :
                case 16 :
                case 22 :
                case 23 :
                case 24 :
                case 25 :
                case 26 : // memo for account
                case 31 : // new mail sound filename
                case 33 : // time to play new mail sound
                case 43 : // smtp user name
                case 47 : // password?
                case 56 : 
                case 61 : // account name (option is 0x04!!)
                case 1000000 : return TRUE;
                case 7 : if (option != 4) { return TRUE; }
        }
        return FALSE;
}

BOOLEAN isNumber(int id, int option) {
        if (option == 4) {
                return TRUE;
        }
        // 6, tw. 7, 9, 12, 13, 14, 17, 18, 19, 20, 21
        return FALSE;
}

BOOLEAN isLastRecord(int id, int option) {
        if (option == -1) {
                return TRUE;
        }
        return FALSE;
}

BOOLEAN isLongNumber(int id, int option) {
        // maybe option means: array of values?
        if (option == 8 && id != 29) {
                return TRUE;
        }
        // 53
        return FALSE;
}

BOOLEAN isCharacter(int id, int option) {
        if (option == 1) {
                return TRUE;
        }
        // 32
        return FALSE;
}

unsigned char *readBitstream(FILE *filehandle, int length) {
        unsigned char *buffer = malloc(length * sizeof(unsigned char));
        int read_bytes = fread(buffer, sizeof(char), length, filehandle);
        if (read_bytes != length) {
                printf("CFG file seems to be corrupt! readBitstream failed (read only %d bytes)\n", read_bytes);
        }
        return buffer;
}

unsigned long readNumber(FILE *filehandle) {
        unsigned char *buffer = readBitstream(filehandle, 4);
        unsigned long number = convert_bitstream_to_number(buffer);
        free(buffer);
        return number;
}

unsigned long long readLongNumber(FILE *filehandle) {
        unsigned char *buffer = readBitstream(filehandle, 8);
        // TODO: Felix - Warum geht hier bitstream to number?
        unsigned long long number = convert_bitstream_to_number(buffer);
        free(buffer);
        return number;
}

unsigned char readCharacter(FILE *filehandle) {
        unsigned char *buffer = readBitstream(filehandle, 1);
        unsigned char character = *buffer;
        return character;
}

void handle_big_binary_block(FILE *filehandle) {
        int jump_length = 1612;
        unsigned char *buffer = readBitstream(filehandle, jump_length);
        int bin_file = open("binary_block.bin", O_CREAT | O_WRONLY | O_TRUNC);
        printf("bin file descriptor should be %d\n", bin_file); 
        write(bin_file, buffer, jump_length);
        close(bin_file);
//        free(buffer);

        printf("skipping binary data (%d)\n", jump_length);
//        fseek(filehandle, jump_length, SEEK_CUR);
}

void readPayload(struct t_record* item, FILE *filehandle) {
        if (isLastRecord(item->id, item->option)) {
                readNumber(filehandle);
                unsigned char *minbuffer = malloc(sizeof(char));
                int read_bytes = fread (minbuffer, sizeof(char), 1, filehandle);
                if (read_bytes != 0 || ! feof(filehandle))    {
                        printf("assertion failed. Should be end of file but isn't!\n");
                }
        }
        else if (isTextfield(item->id, item->option)) {
                item->type = 0;
                int length = readNumber(filehandle);
                printf("Length of data is %d bytes\n", length);
                unsigned char *buffer = malloc((length+1) * sizeof(unsigned char));
                int read_bytes = fread(buffer, sizeof(char), length, filehandle);
                if (read_bytes != length) {
                        printf("CFG file seems to be corrupt! readNumber failed (read only %d bytes)\n", read_bytes);
                        free(buffer);
                }
                else {
                        printf("read text data: %s\n", buffer);
                        bitstream_hexprint(buffer, length);
                        buffer[length] = 0x00;
                        item->data = buffer;
                }
        }
        else if (isNumber(item->id, item->option)) {
                item->type = 1;
                unsigned long *nr = malloc(sizeof(unsigned long));
                *nr = readNumber(filehandle);
                item->data = nr;
        }
        else if (isLongNumber(item->id, item->option)) {
                item->type = 3;
                unsigned long long *nr = malloc(sizeof(unsigned long long));
                *nr = readLongNumber(filehandle);
                item->data = nr;
        }
        else if (item->id == 29 && item->option == 8) {
                // unknown big block! 
                // may consist of multiple records!
                handle_big_binary_block(filehandle);
//                unsigned char *buffer = malloc((20) * sizeof(unsigned char));
//                int read_bytes = fread(buffer, sizeof(char), 20, filehandle);
//                bitstream_hexprint(buffer, 20);
        }
        else if (isCharacter(item->id, item->option)) {
                item->type = 2;
                unsigned char *nr = malloc(sizeof(unsigned char));
                *nr = readCharacter(filehandle);
                item->data = nr;
        }
        else {
                printf("unknown id type %d\n", item->id);
                unsigned char *buffer = malloc((20) * sizeof(unsigned char));
                fread(buffer, sizeof(char), 20, filehandle);
                bitstream_hexprint(buffer, 20);
                exit(0);
        }
}


struct t_record* read_record_start(FILE *filehandle) {
        struct t_record *item = malloc(sizeof(struct t_record));
        item->id = 0;
        item->option = 0;
        item->data = NULL;
        
        item->id = readNumber(filehandle);
        item->option = readNumber(filehandle);
        printf("id is %02x (%02d)\n", item->id, item->id);
        printf("option is %02x\n", item->option);
        return item;
}

BOOLEAN extract_record(FILE *filehandle) {
        struct t_record *item = read_record_start(filehandle);
        if (item != NULL) {
                readPayload(item, filehandle);
                if (item->id == -1) {
                        return FALSE;
                }
        }
        return TRUE;
}


void extract_data(FILE *cfgfile)  {
        while (extract_record(cfgfile) == TRUE)    {
        }
}

int main(int argc, char** argv) {
    char *filename = "ACCOUNT.CFG";
    FILE *cfgfile = fopen(filename, "r");
    if (cfgfile == NULL)    {
        printf("file '%s' could not be opened!\n", filename);
    }
    else {
        extract_data(cfgfile);
        fclose(cfgfile);
    }
    return 0;
}
